/*
                  read energy parameters from a file

                      Stephan Kopp, Ivo Hofacker
                          Vienna RNA Package
*/


#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <math.h>
#include <stdarg.h>
#include "utils.h"
#include "energy_const.h"
#include "energy_par.h"
#include "read_epars.h"

#define PUBLIC
#define PRIVATE   static

#define DEF -50
#define NST 0

/*
#################################
# PRIVATE VARIABLES             #
#################################
*/
PRIVATE FILE *fp;

PRIVATE const char *header = "## RNAfold parameter file v2.0";

PRIVATE const char *identifier[110] = {
  /*
    Do NOT alter the order of the following char pointers!
    They depend on the order of the enum 'parset' as defined
    in read_epars.h, so do NOT reorder the enum as well unless
    you REALLY know what you are doing!
  */
  "END",                                    /* enum type QUIT     */

  /* RNA parameters (normal mode) */
  "stack",                                  /* enum type S        */
  "stack_enthalpies",                       /* enum type S_H      */
  "hairpin",                                /* enum type HP       */
  "hairpin_enthalpies",                     /* enum type HP_H     */
  "bulge",                                  /* enum type B        */
  "bulge_enthalpies",                       /* enum type B_H      */
  "interior",                               /* enum type IL       */
  "interior_enthalpies",                    /* enum type IL_H     */
  "int11",                                  /* enum type INT11    */
  "int11_enthalpies",                       /* enum type INT11_H  */
  "int21",                                  /* enum type INT21    */
  "int21_enthalpies",                       /* enum type INT21_H  */
  "int22",                                  /* enum type INT22    */
  "int22_enthalpies",                       /* enum type INT22_H  */
  "ML_params",                              /* enum type ML       */
  "mismatch_hairpin",                       /* enum type MMH      */
  "mismatch_hairpin_enthalpies",            /* enum type MMH_H    */
  "mismatch_interior",                      /* enum type MMI      */
  "mismatch_interior_enthalpies",           /* enum type MMI_H    */
  "mismatch_interior_1n",                   /* enum type MMI1N    */
  "mismatch_interior_1n_enthalpies",        /* enum type MMI1N_H  */
  "mismatch_interior_23",                   /* enum type MMI23    */
  "mismatch_interior_23_enthalpies",        /* enum type MMI23_H  */
  "mismatch_multi",                         /* enum type MMM      */
  "mismatch_multi_enthalpies",              /* enum type MMM_H    */
  "mismatch_exterior",                      /* enum type MME      */
  "mismatch_exterior_enthalpies",           /* enum type MME_H    */
  "dangle5",                                /* enum type D5       */
  "dangle5_enthalpies",                     /* enum type D5_H     */
  "dangle3",                                /* enum type D3       */
  "dangle3_enthalpies",                     /* enum type D3_H     */
  "Triloops",                               /* enum type TRI      */
  "Tetraloops",                             /* enum type TL       */
  "Hexaloops",                              /* enum type HEX      */
  "NINIO",                                  /* enum type NIN      */
  "Misc",                                   /* enum type MISC     */

  /* DNA parameters for usage in hybrid mode */
  "stack_dna",                              /* enum type S_DNA        */
  "stack_enthalpies_dna",                   /* enum type S_H_DNA      */
  "hairpin_dna",                            /* enum type HP_DNA       */
  "hairpin_enthalpies_dna",                 /* enum type HP_H_DNA     */
  "bulge_dna",                              /* enum type B_DNA        */
  "bulge_enthalpies_dna",                   /* enum type B_H_DNA      */
  "interior_dna",                           /* enum type IL_DNA       */
  "interior_enthalpies_dna",                /* enum type IL_H_DNA     */
  "int11_dna",                              /* enum type INT11_DNA    */
  "int11_enthalpies_dna",                   /* enum type INT11_H_DNA  */
  "int21_dna",                              /* enum type INT21_DNA    */
  "int21_enthalpies_dna",                   /* enum type INT21_H_DNA  */
  "int22_dna",                              /* enum type INT22_DNA    */
  "int22_enthalpies_dna",                   /* enum type INT22_H_DNA  */
  "ML_params_dna",                          /* enum type ML_DNA       */
  "mismatch_hairpin_dna",                   /* enum type MMH_DNA      */
  "mismatch_hairpin_enthalpies_dna",        /* enum type MMH_H_DNA    */
  "mismatch_interior_dna",                  /* enum type MMI_DNA      */
  "mismatch_interior_enthalpies_dna",       /* enum type MMI_H_DNA    */
  "mismatch_interior_1n_dna",               /* enum type MMI1N_DNA    */
  "mismatch_interior_1n_enthalpies_dna",    /* enum type MMI1N_H_DNA  */
  "mismatch_interior_23_dna",               /* enum type MMI23_DNA    */
  "mismatch_interior_23_enthalpies_dna",    /* enum type MMI23_H_DNA  */
  "mismatch_multi_dna",                     /* enum type MMM_DNA      */
  "mismatch_multi_enthalpies_dna",          /* enum type MMM_H_DNA    */
  "mismatch_exterior_dna",                  /* enum type MME_DNA      */
  "mismatch_exterior_enthalpies_dna",       /* enum type MME_H_DNA    */
  "dangle5_dna",                            /* enum type D5_DNA       */
  "dangle5_enthalpies_dna",                 /* enum type D5_H_DNA     */
  "dangle3_dna",                            /* enum type D3_DNA       */
  "dangle3_enthalpies_dna",                 /* enum type D3_H_DNA     */
  "Triloops_dna",                           /* enum type TRI_DNA      */
  "Tetraloops_dna",                         /* enum type TL_DNA       */
  "Hexaloops_dna",                          /* enum type HEX_DNA      */
  "NINIO_dna",                              /* enum type NIN_DNA      */
  "Misc_dna",                               /* enum type MISC_DNA     */

  /* RNA/DNA hybrid parameters for usage in hybrid mode */
  "stack_hybrid",                             /* enum type S_HYBRID        */
  "stack_enthalpies_hybrid",                  /* enum type S_H_HYBRID      */
  "hairpin_hybrid",                           /* enum type HP_HYBRID       */
  "hairpin_enthalpies_hybrid",                /* enum type HP_H_HYBRID     */
  "bulge_hybrid",                             /* enum type B_HYBRID        */
  "bulge_enthalpies_hybrid",                  /* enum type B_H_HYBRID      */
  "interior_hybrid",                          /* enum type IL_HYBRID       */
  "interior_enthalpies_hybrid",               /* enum type IL_H_HYBRID     */
  "int11_hybrid",                             /* enum type INT11_HYBRID    */
  "int11_enthalpies_hybrid",                  /* enum type INT11_H_HYBRID  */
  "int21_hybrid",                             /* enum type INT21_HYBRID    */
  "int21_enthalpies_hybrid",                  /* enum type INT21_H_HYBRID  */
  "int22_hybrid",                             /* enum type INT22_HYBRID    */
  "int22_enthalpies_hybrid",                  /* enum type INT22_H_HYBRID  */
  "ML_params_hybrid",                         /* enum type ML_HYBRID       */
  "mismatch_hairpin_hybrid",                  /* enum type MMH_HYBRID      */
  "mismatch_hairpin_enthalpies_hybrid",       /* enum type MMH_H_HYBRID    */
  "mismatch_interior_hybrid",                 /* enum type MMI_HYBRID      */
  "mismatch_interior_enthalpies_hybrid",      /* enum type MMI_H_HYBRID    */
  "mismatch_interior_1n_hybrid",              /* enum type MMI1N_HYBRID    */
  "mismatch_interior_1n_enthalpies_hybrid",   /* enum type MMI1N_H_HYBRID  */
  "mismatch_interior_23_hybrid",              /* enum type MMI23_HYBRID    */
  "mismatch_interior_23_enthalpies_hybrid",   /* enum type MMI23_H_HYBRID  */
  "mismatch_multi_hybrid",                    /* enum type MMM_HYBRID      */
  "mismatch_multi_enthalpies_hybrid",         /* enum type MMM_H_HYBRID    */
  "mismatch_exterior_hybrid",                 /* enum type MME_HYBRID      */
  "mismatch_exterior_enthalpies_hybrid",      /* enum type MME_H_HYBRID    */
  "dangle5_hybrid",                           /* enum type D5_HYBRID       */
  "dangle5_enthalpies_hybrid",                /* enum type D5_H_HYBRID     */
  "dangle3_hybrid",                           /* enum type D3_HYBRID       */
  "dangle3_enthalpies_hybrid",                /* enum type D3_H_HYBRID     */
  "Triloops_hybrid",                          /* enum type TRI_HYBRID      */
  "Tetraloops_hybrid",                        /* enum type TL_HYBRID       */
  "Hexaloops_hybrid",                         /* enum type HEX_HYBRID      */
  "NINIO_hybrid",                             /* enum type NIN_HYBRID      */
  "Misc_hybrid",                              /* enum type MISC_HYBRID     */

  NULL /* end of char array */
};


/*
#################################
# PRIVATE FUNCTION DECLARATIONS #
#################################
*/

PRIVATE void  display_array(int *p, int size, int line, FILE *fp);

PRIVATE char  *get_array1(int *arr, int size);

PRIVATE void  ignore_comment(char *line);

PRIVATE void  check_symmetry(void);

PRIVATE void  update_nst(int array[NBPAIRS+1][NBPAIRS+1][5][5][5][5]);

/**
 * read a 1dimensional array from file
 * \param array  a pointer to the first element in the array
 * \param dim    the size of the array
 * \param shift  the first position the new values will be written in
 */
PRIVATE void  rd_1dim(int *array, int dim, int shift);

PRIVATE void  rd_1dim_slice(int *array, int dim, int shift, int post);

PRIVATE void  rd_2dim(int *array,
                      int dim1, int dim2,
                      int shift1, int shift2);

PRIVATE void  rd_2dim_slice(int *array,
                      int dim1, int dim2,
                      int shift1, int shift2,
                      int post1, int post2);

PRIVATE void  rd_3dim(int *array,
                      int dim1, int dim2, int dim3,
                      int shift1, int shift2, int shift3);

PRIVATE void  rd_3dim_slice(int *array,
                      int dim1, int dim2, int dim3,
                      int shift1, int shift2, int shift3,
                      int post1, int post2, int post3);

PRIVATE void  rd_4dim(int *array,
                      int dim1, int dim2, int dim3, int dim4,
                      int shift1, int shift2, int shift3, int shift4);

PRIVATE void  rd_4dim_slice(int *array,
                      int dim1, int dim2, int dim3, int dim4,
                      int shift1, int shift2, int shift3, int shift4,
                      int post1, int post2, int post3, int post4);

PRIVATE void  rd_5dim(int *array,
                      int dim1, int dim2, int dim3, int dim4, int dim5,
                      int shift1, int shift2, int shift3, int shift4, int shift5);

PRIVATE void  rd_5dim_slice(int *array,
                      int dim1, int dim2, int dim3, int dim4, int dim5,
                      int shift1, int shift2, int shift3, int shift4, int shift5,
                      int post1, int post2, int post3, int post4, int post5);

PRIVATE void  rd_6dim(int *array,
                      int dim1, int dim2, int dim3, int dim4, int dim5, int dim6,
                      int shift1, int shift2, int shift3, int shift4, int shift5, int shift6);

PRIVATE void  rd_6dim_slice(int *array,
                      int dim1, int dim2, int dim3, int dim4, int dim5, int dim6,
                      int shift1, int shift2, int shift3, int shift4, int shift5, int shift6,
                      int post1, int post2, int post3, int post4, int post5, int post6);

PRIVATE void  rd_SpecialHP(char *seq, int *energy, int *enthalpy, unsigned int size);

/*
#################################
# BEGIN OF FUNCTION DEFINITIONS #
#################################
*/

PUBLIC void read_parameter_file(const char fname[]){
  char        *line, ident[256];
  enum parset type;
  int         r, last;

  last = 0;
  if (!(fp=fopen(fname,"r"))) {
    fprintf(stderr,
            "\nread_parameter_file:\n"
            "\t\tcan't open file %s\n"
            "\t\tusing default parameters instead.\n", fname);
    return;
  }

  if (!(line = get_line(fp))) {
    fprintf(stderr," File %s is inproper.\n", fname);
    fclose(fp);
    return;
  }

  if (strncmp(line, header, 30)!=0) {
    fprintf(stderr,
            "Missing header line in file.\n"
            "Maybe this file has not v2.x format.\n"
            "Use INTERRUPT-key to stop.\n");
  }
  free(line);

  while((line=get_line(fp)) && !last) {

    r = sscanf(line, "# %255s", ident);
    if (r==1) {
      type = gettype(ident);
      switch (type){
        case QUIT:            last = 1;
                              break;
        case S:               rd_2dim(&(stack37[0][0]), NBPAIRS+1, NBPAIRS+1, 1, 1);
                              break;
        case S_H:             rd_2dim(&(stackdH[0][0]), NBPAIRS+1, NBPAIRS+1, 1, 1);
                              break;
        case HP:              rd_1dim(&(hairpin37[0]), 31, 0);
                              break;
        case HP_H:            rd_1dim(&(hairpindH[0]), 31, 0);
                              break;
        case B:               rd_1dim(&(bulge37[0]), 31, 0);
                              break;
        case B_H:             rd_1dim(&(bulgedH[0]), 31, 0);
                              break;
        case IL:              rd_1dim(&(internal_loop37[0]), 31, 0);
                              break;
        case IL_H:            rd_1dim(&(internal_loopdH[0]), 31, 0);
                              break;
        case MME:             rd_3dim(&(mismatchExt37[0][0][0]),
                                  NBPAIRS+1, 5, 5,
                                  1, 0, 0);
                              break;
        case MME_H:           rd_3dim(&(mismatchExtdH[0][0][0]),
                                  NBPAIRS+1, 5, 5,
                                  1, 0, 0);
                              break;
        case MMH:             rd_3dim(&(mismatchH37[0][0][0]),
                                  NBPAIRS+1, 5, 5,
                                  1, 0, 0);
                              break;
        case MMH_H:           rd_3dim(&(mismatchHdH[0][0][0]),
                                  NBPAIRS+1, 5, 5,
                                  1, 0, 0);
                              break;
        case MMI:             rd_3dim(&(mismatchI37[0][0][0]),
                                  NBPAIRS+1, 5, 5,
                                  1, 0, 0);
                              break;
        case MMI_H:           rd_3dim(&(mismatchIdH[0][0][0]),
                                  NBPAIRS+1, 5, 5,
                                  1, 0, 0);
                              break;
        case MMI1N:           rd_3dim(&(mismatch1nI37[0][0][0]),
                                  NBPAIRS+1, 5, 5,
                                  1, 0, 0);
                              break;
        case MMI1N_H:         rd_3dim(&(mismatch1nIdH[0][0][0]),
                                  NBPAIRS+1, 5, 5,
                                  1, 0, 0);
                              break;
        case MMI23:           rd_3dim(&(mismatch23I37[0][0][0]),
                                  NBPAIRS+1, 5, 5,
                                  1, 0, 0);
                              break;
        case MMI23_H:         rd_3dim(&(mismatch23IdH[0][0][0]),
                                  NBPAIRS+1, 5, 5,
                                  1, 0, 0);
                              break;
        case MMM:             rd_3dim(&(mismatchM37[0][0][0]),
                                  NBPAIRS+1, 5, 5,
                                  1, 0, 0);
                              break;
        case MMM_H:           rd_3dim(&(mismatchMdH[0][0][0]),
                                  NBPAIRS+1, 5, 5,
                                  1, 0, 0);
                              break;
        case INT11:           rd_4dim(&(int11_37[0][0][0][0]),
                                  NBPAIRS+1, NBPAIRS+1, 5, 5,
                                  1, 1, 0, 0);
                              break;
        case INT11_H:         rd_4dim(&(int11_dH[0][0][0][0]),
                                  NBPAIRS+1, NBPAIRS+1, 5, 5,
                                  1, 1, 0, 0);
                              break;
        case INT21:           rd_5dim(&(int21_37[0][0][0][0][0]),
                                  NBPAIRS+1, NBPAIRS+1, 5, 5, 5,
                                  1, 1, 0, 0, 0);
                              break;
        case INT21_H:         rd_5dim(&(int21_dH[0][0][0][0][0]),
                                  NBPAIRS+1, NBPAIRS+1, 5, 5, 5,
                                  1, 1, 0, 0, 0);
                              break;
        case INT22:           rd_6dim_slice(&(int22_37[0][0][0][0][0][0]),
                                  NBPAIRS+1, NBPAIRS+1, 5, 5, 5, 5,
                                  1, 1, 1, 1, 1, 1,
                                  1, 1, 0, 0, 0, 0);
                              update_nst(int22_37);
                              break;
        case INT22_H:         rd_6dim_slice(&(int22_dH[0][0][0][0][0][0]),
                                  NBPAIRS+1, NBPAIRS+1, 5, 5, 5, 5,
                                  1, 1, 1, 1, 1, 1,
                                  1, 1, 0, 0, 0, 0);
                              update_nst(int22_dH);
                              break;
        case D5:              rd_2dim(&(dangle5_37[0][0]), NBPAIRS+1, 5, 1, 0);
                              break;
        case D5_H:            rd_2dim(&(dangle5_dH[0][0]), NBPAIRS+1, 5, 1, 0);
                              break;
        case D3:              rd_2dim(&(dangle3_37[0][0]), NBPAIRS+1, 5, 1, 0);
                              break;
        case D3_H:            rd_2dim(&(dangle3_dH[0][0]), NBPAIRS+1, 5, 1, 0);
                              break;
        case ML:              {
                                int values[6];
                                rd_1dim(&values[0], 6, 0);
                                ML_BASE37     = values[0];
                                ML_BASEdH     = values[1];
                                ML_closing37  = values[2];
                                ML_closingdH  = values[3];
                                ML_intern37   = values[4];
                                ML_interndH   = values[5];
                              }
                              break;
        case NIN:             {
                                int values[3];
                                rd_1dim(&values[0], 3, 0);
                                ninio37 = values[0];
                                niniodH = values[1];
                                MAX_NINIO  = values[2];
                              }
                              break;
        case MISC:            {
                                int values[4];
                                rd_1dim(&values[0], 4, 0);
                                DuplexInit37 = values[0];
                                DuplexInitdH = values[1];
                                TerminalAU37 = values[2];
                                TerminalAUdH = values[3];
                              }
                              break;
        case TL:              rd_SpecialHP(&(Tetraloops[0]), &(Tetraloop37[0]), &(TetraloopdH[0]), 4);
                              break;
        case TRI:             rd_SpecialHP(&(Triloops[0]), &(Triloop37[0]), &(TriloopdH[0]), 3);
                              break;
        case HEX:             rd_SpecialHP(&(Hexaloops[0]), &(Hexaloop37[0]), &(HexaloopdH[0]), 6);
                              break;

        case S_DNA:           rd_2dim(&(stack37_D[0][0]), NBPAIRS+1, NBPAIRS+1, 1, 1);
                              break;
        case S_H_DNA:         rd_2dim(&(stackdH_D[0][0]), NBPAIRS+1, NBPAIRS+1, 1, 1);
                              break;
        case HP_DNA:          rd_1dim(&(hairpin37_D[0]), 31, 0);
                              break;
        case HP_H_DNA:        rd_1dim(&(hairpindH_D[0]), 31, 0);
                              break;
        case B_DNA:           rd_1dim(&(bulge37_D[0]), 31, 0);
                              break;
        case B_H_DNA:         rd_1dim(&(bulgedH_D[0]), 31, 0);
                              break;
        case IL_DNA:          rd_1dim(&(internal_loop37_D[0]), 31, 0);
                              break;
        case IL_H_DNA:        rd_1dim(&(internal_loopdH_D[0]), 31, 0);
                              break;
        case MME_DNA:         rd_3dim(&(mismatchExt37_D[0][0][0]),
                                      NBPAIRS+1, 5, 5,
                                      1, 0, 0);
                              break;
        case MME_H_DNA:       rd_3dim(&(mismatchExtdH_D[0][0][0]),
                                      NBPAIRS+1, 5, 5,
                                      1, 0, 0);
                              break;
        case MMH_DNA:         rd_3dim(&(mismatchH37_D[0][0][0]),
                                      NBPAIRS+1, 5, 5,
                                      1, 0, 0);
                              break;
        case MMH_H_DNA:       rd_3dim(&(mismatchHdH_D[0][0][0]),
                                      NBPAIRS+1, 5, 5,
                                      1, 0, 0);
                              break;
        case MMI_DNA:         rd_3dim(&(mismatchI37_D[0][0][0]),
                                      NBPAIRS+1, 5, 5,
                                      1, 0, 0);
                              break;
        case MMI_H_DNA:       rd_3dim(&(mismatchIdH_D[0][0][0]),
                                      NBPAIRS+1, 5, 5,
                                      1, 0, 0);
                              break;
        case MMI1N_DNA:       rd_3dim(&(mismatch1nI37_D[0][0][0]),
                                      NBPAIRS+1, 5, 5,
                                      1, 0, 0);
                              break;
        case MMI1N_H_DNA:     rd_3dim(&(mismatch1nIdH_D[0][0][0]),
                                      NBPAIRS+1, 5, 5,
                                      1, 0, 0);
                              break;
        case MMI23_DNA:       rd_3dim(&(mismatch23I37_D[0][0][0]),
                                      NBPAIRS+1, 5, 5,
                                      1, 0, 0);
                              break;
        case MMI23_H_DNA:     rd_3dim(&(mismatch23IdH_D[0][0][0]),
                                      NBPAIRS+1, 5, 5,
                                      1, 0, 0);
                              break;
        case MMM_DNA:         rd_3dim(&(mismatchM37_D[0][0][0]),
                                      NBPAIRS+1, 5, 5,
                                      1, 0, 0);
                              break;
        case MMM_H_DNA:       rd_3dim(&(mismatchMdH_D[0][0][0]),
                                      NBPAIRS+1, 5, 5,
                                      1, 0, 0);
                              break;
        case INT11_DNA:       rd_4dim(&(int11_37_D[0][0][0][0]),
                                      NBPAIRS+1, NBPAIRS+1, 5, 5,
                                      1, 1, 0, 0);
                              break;
        case INT11_H_DNA:     rd_4dim(&(int11_dH_D[0][0][0][0]),
                                      NBPAIRS+1, NBPAIRS+1, 5, 5,
                                      1, 1, 0, 0);
                              break;
        case INT21_DNA:       rd_5dim(&(int21_37_D[0][0][0][0][0]),
                                      NBPAIRS+1, NBPAIRS+1, 5, 5, 5,
                                      1, 1, 0, 0, 0);
                              break;
        case INT21_H_DNA:     rd_5dim(&(int21_dH_D[0][0][0][0][0]),
                                      NBPAIRS+1, NBPAIRS+1, 5, 5, 5,
                                      1, 1, 0, 0, 0);
                              break;
        case INT22_DNA:       rd_6dim_slice(&(int22_37_D[0][0][0][0][0][0]),
                                            NBPAIRS+1, NBPAIRS+1, 5, 5, 5, 5,
                                            1, 1, 1, 1, 1, 1,
                                            1, 1, 0, 0, 0, 0);
                              update_nst(int22_37_D);
                              break;
        case INT22_H_DNA:     rd_6dim_slice(&(int22_dH_D[0][0][0][0][0][0]),
                                            NBPAIRS+1, NBPAIRS+1, 5, 5, 5, 5,
                                            1, 1, 1, 1, 1, 1,
                                            1, 1, 0, 0, 0, 0);
                              update_nst(int22_dH_D);
                              break;
        case D5_DNA:          rd_2dim(&(dangle5_37_D[0][0]), NBPAIRS+1, 5, 1, 0);
                              break;
        case D5_H_DNA:        rd_2dim(&(dangle5_dH_D[0][0]), NBPAIRS+1, 5, 1, 0);
                              break;
        case D3_DNA:          rd_2dim(&(dangle3_37_D[0][0]), NBPAIRS+1, 5, 1, 0);
                              break;
        case D3_H_DNA:        rd_2dim(&(dangle3_dH_D[0][0]), NBPAIRS+1, 5, 1, 0);
                              break;
        case ML_DNA:          {
                                int values[6];
                                rd_1dim(&values[0], 6, 0);
                                ML_BASE37_D     = values[0];
                                ML_BASEdH_D     = values[1];
                                ML_closing37_D  = values[2];
                                ML_closingdH_D  = values[3];
                                ML_intern37_D   = values[4];
                                ML_interndH_D   = values[5];
                              }
                              break;
        case NIN_DNA:         {
                                int values[3];
                                rd_1dim(&values[0], 3, 0);
                                ninio37_D = values[0];
                                niniodH_D = values[1];
                                MAX_NINIO_D  = values[2];
                              }
                              break;
        case MISC_DNA:        {
                                int values[4];
                                rd_1dim(&values[0], 4, 0);
                                DuplexInit37_D = values[0];
                                DuplexInitdH_D = values[1];
                                TerminalAU37_D = values[2];
                                TerminalAUdH_D = values[3];
                              }
                              break;
        case TL_DNA:          rd_SpecialHP(&(Tetraloops_D[0]), &(Tetraloop37_D[0]), &(TetraloopdH_D[0]), 4);
                              break;
        case TRI_DNA:         rd_SpecialHP(&(Triloops_D[0]), &(Triloop37_D[0]), &(TriloopdH_D[0]), 3);
                              break;
        case HEX_DNA:         rd_SpecialHP(&(Hexaloops_D[0]), &(Hexaloop37_D[0]), &(HexaloopdH_D[0]), 6);
                              break;

        case S_HYBRID:        rd_2dim(&(stack37_RD[0][0]), NBPAIRS+1, NBPAIRS+1, 1, 1);
                              break;
        case S_H_HYBRID:      rd_2dim(&(stackdH_RD[0][0]), NBPAIRS+1, NBPAIRS+1, 1, 1);
                              break;
        case HP_HYBRID:       rd_1dim(&(hairpin37_RD[0]), 31, 0);
                              break;
        case HP_H_HYBRID:     rd_1dim(&(hairpindH_RD[0]), 31, 0);
                              break;
        case B_HYBRID:        rd_1dim(&(bulge37_RD[0]), 31, 0);
                              break;
        case B_H_HYBRID:      rd_1dim(&(bulgedH_RD[0]), 31, 0);
                              break;
        case IL_HYBRID:       rd_1dim(&(internal_loop37_RD[0]), 31, 0);
                              break;
        case IL_H_HYBRID:     rd_1dim(&(internal_loopdH_RD[0]), 31, 0);
                              break;
        case MME_HYBRID:      rd_3dim(&(mismatchExt37_RD[0][0][0]),
                                      NBPAIRS+1, 5, 5,
                                      1, 0, 0);
                              break;
        case MME_H_HYBRID:    rd_3dim(&(mismatchExtdH_RD[0][0][0]),
                                      NBPAIRS+1, 5, 5,
                                      1, 0, 0);
                              break;
        case MMH_HYBRID:      rd_3dim(&(mismatchH37_RD[0][0][0]),
                                      NBPAIRS+1, 5, 5,
                                      1, 0, 0);
                              break;
        case MMH_H_HYBRID:    rd_3dim(&(mismatchHdH_RD[0][0][0]),
                                      NBPAIRS+1, 5, 5,
                                      1, 0, 0);
                              break;
        case MMI_HYBRID:      rd_3dim(&(mismatchI37_RD[0][0][0]),
                                      NBPAIRS+1, 5, 5,
                                      1, 0, 0);
                              break;
        case MMI_H_HYBRID:    rd_3dim(&(mismatchIdH_RD[0][0][0]),
                                      NBPAIRS+1, 5, 5,
                                      1, 0, 0);
                              break;
        case MMI1N_HYBRID:    rd_3dim(&(mismatch1nI37_RD[0][0][0]),
                                      NBPAIRS+1, 5, 5,
                                      1, 0, 0);
                              break;
        case MMI1N_H_HYBRID:  rd_3dim(&(mismatch1nIdH_RD[0][0][0]),
                                      NBPAIRS+1, 5, 5,
                                      1, 0, 0);
                              break;
        case MMI23_HYBRID:    rd_3dim(&(mismatch23I37_RD[0][0][0]),
                                      NBPAIRS+1, 5, 5,
                                      1, 0, 0);
                              break;
        case MMI23_H_HYBRID:  rd_3dim(&(mismatch23IdH_RD[0][0][0]),
                                      NBPAIRS+1, 5, 5,
                                      1, 0, 0);
                              break;
        case MMM_HYBRID:      rd_3dim(&(mismatchM37_RD[0][0][0]),
                                      NBPAIRS+1, 5, 5,
                                      1, 0, 0);
                              break;
        case MMM_H_HYBRID:    rd_3dim(&(mismatchMdH_RD[0][0][0]),
                                      NBPAIRS+1, 5, 5,
                                      1, 0, 0);
                              break;
        case INT11_HYBRID:    rd_4dim(&(int11_37_RD[0][0][0][0]),
                                      NBPAIRS+1, NBPAIRS+1, 5, 5,
                                      1, 1, 0, 0);
                              break;
        case INT11_H_HYBRID:  rd_4dim(&(int11_dH_RD[0][0][0][0]),
                                      NBPAIRS+1, NBPAIRS+1, 5, 5,
                                      1, 1, 0, 0);
                              break;
        case INT21_HYBRID:    rd_5dim(&(int21_37_RD[0][0][0][0][0]),
                                      NBPAIRS+1, NBPAIRS+1, 5, 5, 5,
                                      1, 1, 0, 0, 0);
                              break;
        case INT21_H_HYBRID:  rd_5dim(&(int21_dH_RD[0][0][0][0][0]),
                                      NBPAIRS+1, NBPAIRS+1, 5, 5, 5,
                                      1, 1, 0, 0, 0);
                              break;
        case INT22_HYBRID:    rd_6dim_slice(&(int22_37_RD[0][0][0][0][0][0]),
                                            NBPAIRS+1, NBPAIRS+1, 5, 5, 5, 5,
                                            1, 1, 1, 1, 1, 1,
                                            1, 1, 0, 0, 0, 0);
                              update_nst(int22_37_RD);
                              break;
        case INT22_H_HYBRID:  rd_6dim_slice(&(int22_dH_RD[0][0][0][0][0][0]),
                                            NBPAIRS+1, NBPAIRS+1, 5, 5, 5, 5,
                                            1, 1, 1, 1, 1, 1,
                                            1, 1, 0, 0, 0, 0);
                              update_nst(int22_dH_RD);
                              break;
        case D5_HYBRID:       rd_2dim(&(dangle5_37_RD[0][0]), NBPAIRS+1, 5, 1, 0);
                              break;
        case D5_H_HYBRID:     rd_2dim(&(dangle5_dH_RD[0][0]), NBPAIRS+1, 5, 1, 0);
                              break;
        case D3_HYBRID:       rd_2dim(&(dangle3_37_RD[0][0]), NBPAIRS+1, 5, 1, 0);
                              break;
        case D3_H_HYBRID:     rd_2dim(&(dangle3_dH_RD[0][0]), NBPAIRS+1, 5, 1, 0);
                              break;
        case ML_HYBRID:       {
                                int values[6];
                                rd_1dim(&values[0], 6, 0);
                                ML_BASE37_RD     = values[0];
                                ML_BASEdH_RD     = values[1];
                                ML_closing37_RD  = values[2];
                                ML_closingdH_RD  = values[3];
                                ML_intern37_RD   = values[4];
                                ML_interndH_RD   = values[5];
                              }
                              break;
        case NIN_HYBRID:      {
                                int values[3];
                                rd_1dim(&values[0], 3, 0);
                                ninio37_RD = values[0];
                                niniodH_RD = values[1];
                                MAX_NINIO_RD  = values[2];
                              }
                              break;
        case MISC_HYBRID:     {
                                int values[4];
                                rd_1dim(&values[0], 4, 0);
                                DuplexInit37_RD = values[0];
                                DuplexInitdH_RD = values[1];
                                TerminalAU37_RD = values[2];
                                TerminalAUdH_RD = values[3];
                              }
                              break;
        case TL_HYBRID:       rd_SpecialHP(&(Tetraloops_RD[0]), &(Tetraloop37_RD[0]), &(TetraloopdH_RD[0]), 4);
                              break;
        case TRI_HYBRID:      rd_SpecialHP(&(Triloops_RD[0]), &(Triloop37_RD[0]), &(TriloopdH_RD[0]), 3);
                              break;
        case HEX_HYBRID:      rd_SpecialHP(&(Hexaloops_RD[0]), &(Hexaloop37_RD[0]), &(HexaloopdH_RD[0]), 6);
                              break;

        default:          /* do nothing but complain */
                          fprintf(stderr,"read_epars: Unknown field identifier in `%s'\n", line);
      }
    } /* else ignore line */
    free(line);
  }
  fclose(fp);

  check_symmetry();
  return;
}

/*------------------------------------------------------------*/

PRIVATE void display_array(int *p, int size, int nl, FILE *fp)
{
  int i;

  for (i=1; i<=size; i++, p++) {
    switch(*p)
      {
      case  INF: fprintf(fp,"   INF"); break;
      case -INF: fprintf(fp,"  -INf"); break;
      case  DEF: fprintf(fp,"   DEF"); break;
      default:   fprintf(fp,"%6d",  *p); break;
      }
    if ((i%nl)==0) fprintf(fp,"\n");
  }
  if (size%nl) fprintf(fp,"\n");
  return;
}

/*------------------------------------------------------------*/

PRIVATE char *get_array1(int *arr, int size)
{
  int    i, p, pos, pp, r, last;
  char  *line, buf[16];


  i = last = 0;
  while( i<size ) {
    line = get_line(fp);
    if (!line) nrerror("unexpected end of file in get_array1");
    ignore_comment(line);
    pos=0;
    while ((i<size)&&(sscanf(line+pos,"%15s%n", buf, &pp)==1)) {
      pos += pp;
      if (buf[0]=='*') {i++; continue;}
      else if (buf[0]=='x') { /* should only be used for loop parameters */
        if (i==0) nrerror("can't extrapolate first value");
        p = arr[last] + (int) (0.5+ lxc37*log(((double) i)/(double)(last)));
      }
      else if (strcmp(buf,"DEF") == 0) p = DEF;
      else if (strcmp(buf,"INF") == 0) p = INF;
      else if (strcmp(buf,"NST") == 0) p = NST;
      else {
        r=sscanf(buf,"%d", &p);
        if (r!=1) {
          return line+pos;
          fprintf(stderr, "can't interpret `%s' in get_array1\n", buf);
          exit(1);
        }
        last = i;
      }
      arr[i++]=p;
    }
    free(line);
  }

  return NULL;
}

PRIVATE void rd_1dim( int *array,
                      int dim,
                      int shift){

  rd_1dim_slice(array, dim, shift, 0);
}

PRIVATE void rd_1dim_slice( int *array,
                            int dim,
                            int shift,
                            int post){

  char *cp;
  cp   = get_array1(array+shift, dim-shift-post);

  if (cp) {
    fprintf(stderr,"\nrd_1dim: %s\n", cp);
    exit(1);
  }
  return;
}

PRIVATE void  rd_2dim(int *array,
                      int dim1, int dim2,
                      int shift1, int shift2){

  rd_2dim_slice(array, dim1, dim2, shift1, shift2, 0, 0);
}

PRIVATE void  rd_2dim_slice(int *array,
                            int dim1, int dim2,
                            int shift1, int shift2,
                            int post1, int post2){

  int i;
  int delta_pre   = shift1 + shift2;
  int delta_post  = post1 + post2;

  if(delta_pre + delta_post == 0){
    rd_1dim(array, dim1 * dim2, 0);
    return;
  }
  for (i=shift1; i<dim1 - post1; i++)
    rd_1dim_slice(array + (i*dim2), dim2, shift2, post2);
  return;
}

PRIVATE void  rd_3dim(int *array,
                      int dim1, int dim2, int dim3,
                      int shift1, int shift2, int shift3){

  rd_3dim_slice(array,
                dim1, dim2, dim3,
                shift1, shift2, shift3,
                0, 0, 0);
}

PRIVATE void  rd_3dim_slice(int *array,
                            int dim1, int dim2, int dim3,
                            int shift1, int shift2, int shift3,
                            int post1, int post2, int post3){

  int    i;
  int delta_pre   = shift1 + shift2 + shift3;
  int delta_post  = post1 + post2 + post3;

  if(delta_pre + delta_post == 0){
    rd_1dim(array, dim1 * dim2 * dim3, 0);
    return;
  }
  for (i=shift1; i<dim1 - post1; i++){
    rd_2dim_slice(array + (i * dim2 * dim3),
            dim2, dim3,
            shift2, shift3,
            post2, post3);
  }
  return;
}

PRIVATE void  rd_4dim(int *array,
                      int dim1, int dim2, int dim3, int dim4,
                      int shift1, int shift2, int shift3, int shift4){

  rd_4dim_slice(array,
                dim1, dim2, dim3, dim4,
                shift1, shift2, shift3, shift4,
                0, 0, 0, 0);
}

PRIVATE void  rd_4dim_slice(int *array,
                      int dim1, int dim2, int dim3, int dim4,
                      int shift1, int shift2, int shift3, int shift4,
                      int post1, int post2, int post3, int post4){

  int i;
  int delta_pre   = shift1 + shift2 + shift3 + shift4;
  int delta_post  = post1 + post2 + post3 + post4;

  if(delta_pre + delta_post == 0){
    rd_1dim(array, dim1 * dim2 * dim3 * dim4, 0);
    return;
  }
  for(i=shift1; i<dim1 - post1; i++){
    rd_3dim_slice(array + (i * dim2 * dim3 * dim4),
            dim2, dim3, dim4,
            shift2, shift3, shift4,
            post2, post3, post4);
  }
  return;
}

PRIVATE void  rd_5dim(int *array,
                      int dim1, int dim2, int dim3, int dim4, int dim5,
                      int shift1, int shift2, int shift3, int shift4, int shift5){

  rd_5dim_slice(array,
                dim1, dim2, dim3, dim4, dim5,
                shift1, shift2, shift3, shift4, shift5,
                0, 0, 0, 0, 0);
}

PRIVATE void  rd_5dim_slice(int *array,
                      int dim1, int dim2, int dim3, int dim4, int dim5,
                      int shift1, int shift2, int shift3, int shift4, int shift5,
                      int post1, int post2, int post3, int post4, int post5){

  int i;
  int delta_pre   = shift1 + shift2 + shift3 + shift4 + shift5;
  int delta_post  = post1 + post2 + post3 + post4 + post5;

  if(delta_pre + delta_post == 0){
    rd_1dim(array, dim1 * dim2 * dim3 * dim4 * dim5, 0);
    return;
  }
  for(i=shift1; i<dim1 - post1; i++)
    rd_4dim_slice(array + (i * dim2 * dim3 * dim4 * dim5),
            dim2, dim3, dim4, dim5,
            shift2, shift3, shift4, shift5,
            post2, post3, post4, post5);
  return;
}

/**
*** \param dim1   The size of the first dimension
*** \param shift1 The pre shift for the first dimension
**/
PRIVATE void  rd_6dim(int *array,
                      int dim1, int dim2, int dim3, int dim4, int dim5, int dim6,
                      int shift1, int shift2, int shift3, int shift4, int shift5, int shift6){

  rd_6dim_slice(array,
                dim1, dim2, dim3, dim4, dim5, dim6,
                shift1, shift2, shift3, shift4, shift5, shift6,
                0, 0, 0, 0, 0, 0);
}

PRIVATE void  rd_6dim_slice(int *array,
                      int dim1, int dim2, int dim3, int dim4, int dim5, int dim6,
                      int shift1, int shift2, int shift3, int shift4, int shift5, int shift6,
                      int post1, int post2, int post3, int post4, int post5, int post6){

  int i;
  int delta_pre   = shift1 + shift2 + shift3 + shift4 + shift5 + shift6;
  int delta_post  = post1 + post2 + post3 + post4 + post5 + post6;

  if(delta_pre + delta_post == 0){
    rd_1dim(array, dim1 * dim2 * dim3 * dim4 * dim5 * dim6, 0);
    return;
  }
  for(i=shift1; i<dim1 - post1; i++)
    rd_5dim_slice(array + (i * dim2 * dim3 * dim4 * dim5 * dim6),
            dim2, dim3, dim4, dim5, dim6,
            shift2, shift3, shift4, shift5, shift6,
            post2, post3, post4, post5, post6);
  return;
}


PRIVATE void  rd_SpecialHP(char *seq, int *energy, int *enthalpy, unsigned int size){
  int    i, r;
  char   *buf, format_string[10];

  if(size > 6) nrerror("rd_SpecialHP(): size of HP too large");

  i = 0;

  /* erase old tetraloop entries */
  memset(seq, 0, 40 * (size + 3) + 1);
  memset(energy, 0, 40 * sizeof(int));
  memset(enthalpy, 0, 40 * sizeof(int));

  sprintf(&(format_string[0]), "%%%ds %%d %%d", size+2);

  do {
    buf = get_line(fp);
    if (buf==NULL) break;
    r = sscanf(buf, format_string, seq + ((size + 3) * i), energy + i, enthalpy + i);
    strcat(seq, " ");
    free(buf);
    i++;
  } while((r==3)&&(i<40));
  return;
}


PRIVATE void ignore_comment(char * line)
{
  /* excise C style comments */
  /* only one comment per line, no multiline comments */
  char *cp1, *cp2;

  if ((cp1=strstr(line, "/*"))) {
    cp2 = strstr(cp1, "*/");
    if (cp2==NULL)
      nrerror("unclosed comment in parameter file");
    /* can't use strcpy for overlapping strings */
    for (cp2+=2; *cp2!='\0'; cp2++, cp1++)
      *cp1 = *cp2;
    *cp1 = '\0';
  }

  return;
}
/*------------------------------------------------------------*/

PUBLIC const char *settype(enum parset s){
  if((s >= QUIT) && (s < DUMMY))
    return identifier[s];
  else
    nrerror("\nThe answer is: 42\n");
}

/*------------------------------------------------------------*/
PUBLIC enum parset gettype(const char *ident){
  int i = 0;
  while(identifier[i]){
    if(strcmp(ident, identifier[i]) == 0) return i;
    i++;
  }
  return UNKNOWN;
}

/*---------------------------------------------------------------*/

PUBLIC void write_parameter_file(const char fname[]){
  write_parameter_file_full(fname, VRNA_PARAMETER_FILE_RNA);
}

PUBLIC void write_parameter_file_full(const char fname[], int options){
  FILE *outfp;
  int c;
  const char *pnames[]          = {"NP", "CG", "GC", "GU", "UG", "AU", "UA", " @"};
  const char bnames[]           = "@ACGU";
  const char *format_identifier = "\n# %s\n";
  const char *format_pp         = "/* %2s..%2s */\n";
  const char *format_pbp        = "/* %2s.%c..%2s */\n";
  const char *format_pbbp       = "/* %2s.%c%c..%2s */\n";
  const char *pairs_comment     = "/*  CG    GC    GU    UG    AU    UA    @  */\n";
  const char *bases_comment     = "/*  @     A     C     G     U   */\n";

  outfp = fopen(fname, "w");
  if (!outfp) {
    fprintf(stderr, "can't open file %s\n", fname);
    exit(1);
  }
  fprintf(outfp, "%s\n", header);

  if(options & VRNA_PARAMETER_FILE_RNA){
    fprintf(outfp, format_identifier, settype(S));
    fprintf(outfp, pairs_comment);
    for (c=1; c<NBPAIRS+1; c++)
      display_array(stack37[c]+1,NBPAIRS,NBPAIRS, outfp);

    fprintf(outfp, format_identifier, settype(S_H));
    fprintf(outfp, pairs_comment);
    for (c=1; c<NBPAIRS+1; c++)
      display_array(stackdH[c]+1,NBPAIRS,NBPAIRS, outfp);

    fprintf(outfp, format_identifier, settype(MMH));
    { int i,k;
    for (k=1; k<NBPAIRS+1; k++)
      for (i=0; i<5; i++)
        display_array(mismatchH37[k][i],5,5, outfp);
    }

    fprintf(outfp, format_identifier, settype(MMH_H));
    { int i,k;
    for (k=1; k<NBPAIRS+1; k++)
      for (i=0; i<5; i++)
        display_array(mismatchHdH[k][i],5,5, outfp);
    }

    fprintf(outfp, format_identifier, settype(MMI));
    { int i,k;
    for (k=1; k<NBPAIRS+1; k++)
      for (i=0; i<5; i++)
        display_array(mismatchI37[k][i],5,5, outfp);
    }

    fprintf(outfp, format_identifier, settype(MMI_H));
    { int i,k;
    for (k=1; k<NBPAIRS+1; k++)
      for (i=0; i<5; i++)
        display_array(mismatchIdH[k][i],5,5, outfp);
    }

    fprintf(outfp, format_identifier, settype(MMI1N));
    { int i,k;
    for (k=1; k<NBPAIRS+1; k++)
      for (i=0; i<5; i++)
        display_array(mismatch1nI37[k][i],5,5, outfp);
    }

    fprintf(outfp, format_identifier, settype(MMI1N_H));
    { int i,k;
    for (k=1; k<NBPAIRS+1; k++)
      for (i=0; i<5; i++)
        display_array(mismatch1nIdH[k][i],5,5, outfp);
    }

    fprintf(outfp, format_identifier, settype(MMI23));
    { int i,k;
    for (k=1; k<NBPAIRS+1; k++)
      for (i=0; i<5; i++)
        display_array(mismatch23I37[k][i],5,5, outfp);
    }

    fprintf(outfp, format_identifier, settype(MMI23_H));
    { int i,k;
    for (k=1; k<NBPAIRS+1; k++)
      for (i=0; i<5; i++)
        display_array(mismatch23IdH[k][i],5,5, outfp);
    }

    fprintf(outfp, format_identifier, settype(MMM));
    { int i,k;
    for (k=1; k<NBPAIRS+1; k++)
      for (i=0; i<5; i++)
        display_array(mismatchM37[k][i],5,5, outfp);
    }

    fprintf(outfp, format_identifier, settype(MMM_H));
    { int i,k;
    for (k=1; k<NBPAIRS+1; k++)
      for (i=0; i<5; i++)
        display_array(mismatchMdH[k][i],5,5, outfp);
    }

    fprintf(outfp, format_identifier, settype(MME));
    { int i,k;
    for (k=1; k<NBPAIRS+1; k++)
      for (i=0; i<5; i++)
        display_array(mismatchExt37[k][i],5,5, outfp);
    }

    fprintf(outfp, format_identifier, settype(MME_H));
    { int i,k;
    for (k=1; k<NBPAIRS+1; k++)
      for (i=0; i<5; i++)
        display_array(mismatchExtdH[k][i],5,5, outfp);
    }

    fprintf(outfp, format_identifier, settype(D5));
    fprintf(outfp, bases_comment);
    for (c=1; c<NBPAIRS+1; c++)
      display_array(dangle5_37[c], 5, 5, outfp);

    fprintf(outfp, format_identifier, settype(D5_H));
    fprintf(outfp, bases_comment);
    for (c=1; c<NBPAIRS+1; c++)
      display_array(dangle5_dH[c], 5, 5, outfp);

    fprintf(outfp, format_identifier, settype(D3));
    fprintf(outfp, bases_comment);
    for (c=1; c<NBPAIRS+1; c++)
      display_array(dangle3_37[c], 5, 5, outfp);

    fprintf(outfp, format_identifier, settype(D3_H));
    fprintf(outfp, bases_comment);
    for (c=1; c<NBPAIRS+1; c++)
      display_array(dangle3_dH[c], 5, 5, outfp);


    /* dont print "no pair" entries for interior loop arrays */
    fprintf(outfp, format_identifier, settype(INT11));
    { int i,k,l;
    for (k=1; k<NBPAIRS+1; k++)
      for (l=1; l<NBPAIRS+1; l++) {
        fprintf(outfp, format_pp, pnames[k], pnames[l]);
        for (i=0; i<5; i++)
          display_array(int11_37[k][l][i], 5, 5, outfp);
      }
    }

    fprintf(outfp, format_identifier, settype(INT11_H));
    { int i,k,l;
    for (k=1; k<NBPAIRS+1; k++)
      for (l=1; l<NBPAIRS+1; l++) {
        fprintf(outfp, format_pp, pnames[k], pnames[l]);
        for (i=0; i<5; i++)
          display_array(int11_dH[k][l][i],5,5, outfp);
      }
    }

    fprintf(outfp, format_identifier, settype(INT21));
    { int p1, p2, i, j;
    for (p1=1; p1<NBPAIRS+1; p1++)
      for (p2=1; p2<NBPAIRS+1; p2++)
        for (i=0; i<5; i++) {
          fprintf(outfp, format_pbp,
                  pnames[p1], bnames[i], pnames[p2]);
          for (j=0; j<5; j++)
            display_array(int21_37[p1][p2][i][j],5,5, outfp);
        }
    }

    fprintf(outfp, format_identifier, settype(INT21_H));
    { int p1, p2, i, j;
    for (p1=1; p1<NBPAIRS+1; p1++)
      for (p2=1; p2<NBPAIRS+1; p2++)
        for (i=0; i<5; i++) {
          fprintf(outfp, format_pbp,
                  pnames[p1], bnames[i], pnames[p2]);
          for (j=0; j<5; j++)
            display_array(int21_dH[p1][p2][i][j],5,5, outfp);
        }
    }

    fprintf(outfp, format_identifier, settype(INT22));
    { int p1, p2, i, j, k;
    for (p1=1; p1<NBPAIRS; p1++)
      for (p2=1; p2<NBPAIRS; p2++)
        for (i=1; i<5; i++)
          for (j=1; j<5; j++) {
            fprintf(outfp, format_pbbp,
                    pnames[p1], bnames[i], bnames[j], pnames[p2]);
            for (k=1; k<5; k++)
              display_array(int22_37[p1][p2][i][j][k]+1,4,5, outfp);
          }
    }

    fprintf(outfp, format_identifier, settype(INT22_H));
    { int p1, p2, i, j, k;
    for (p1=1; p1<NBPAIRS; p1++)
      for (p2=1; p2<NBPAIRS; p2++)
        for (i=1; i<5; i++)
          for (j=1; j<5; j++) {
            fprintf(outfp, format_pbbp,
                    pnames[p1], bnames[i], bnames[j], pnames[p2]);
            for (k=1; k<5; k++)
              display_array(int22_dH[p1][p2][i][j][k]+1,4,5, outfp);
          }
    }

    fprintf(outfp, format_identifier, settype(HP));
    display_array(hairpin37, 31, 10, outfp);

    fprintf(outfp, format_identifier, settype(HP_H));
    display_array(hairpindH, 31, 10, outfp);

    fprintf(outfp, format_identifier, settype(B));
    display_array(bulge37, 31, 10, outfp);

    fprintf(outfp, format_identifier, settype(B_H));
    display_array(bulgedH, 31, 10, outfp);

    fprintf(outfp, format_identifier, settype(IL));
    display_array(internal_loop37, 31, 10, outfp);

    fprintf(outfp, format_identifier, settype(IL_H));
    display_array(internal_loopdH, 31, 10, outfp);

    fprintf(outfp, format_identifier, settype(ML));
    fprintf(outfp,"/* F = cu*n_unpaired + cc + ci*loop_degree (+TermAU) */\n");
    fprintf(outfp,"/*\t    cu\t cu_dH\t    cc\t cc_dH\t    ci\t ci_dH  */\n");
    fprintf(outfp,"\t%6d\t%6d\t%6d\t%6d\t%6d\t%6d\n", ML_BASE37, ML_BASEdH, ML_closing37, ML_closingdH, ML_intern37, ML_interndH);

    fprintf(outfp, format_identifier, settype(NIN));
    fprintf(outfp,"/* Ninio = MIN(max, m*|n1-n2| */\n"
                "/*\t    m\t  m_dH     max  */\n"
                "\t%6d\t%6d\t%6d\n", ninio37, niniodH, MAX_NINIO);

    fprintf(outfp, format_identifier, settype(MISC));
    fprintf(outfp,"/* all parameters are pairs of 'energy enthalpy' */\n");
    fprintf(outfp,"/*    DuplexInit     TerminalAU      LXC */\n");
    fprintf(outfp,"   %6d %6d %6d  %6d %3.6f %6d\n", DuplexInit37, DuplexInitdH, TerminalAU37, TerminalAUdH, lxc37, 0);

    fprintf(outfp, format_identifier, settype(HEX));
    for (c=0; c< strlen(Hexaloops)/9; c++)
      fprintf(outfp,"\t%.8s %6d %6d\n", Hexaloops+c*9, Hexaloop37[c], HexaloopdH[c]);

    fprintf(outfp, format_identifier, settype(TL));
    for (c=0; c< strlen(Tetraloops)/7; c++)
      fprintf(outfp,"\t%.6s %6d %6d\n", Tetraloops+c*7, Tetraloop37[c], TetraloopdH[c]);

    fprintf(outfp, format_identifier, settype(TRI));
    for (c=0; c< strlen(Triloops)/6; c++)
      fprintf(outfp,"\t%.5s %6d %6d\n", Triloops+c*6, Triloop37[c], TriloopdH[c]);
  }

  if(options & VRNA_PARAMETER_FILE_DNA){
    fprintf(outfp, format_identifier, settype(S_DNA));
    fprintf(outfp, pairs_comment);
    for (c=1; c<NBPAIRS+1; c++)
      display_array(stack37_D[c]+1,NBPAIRS,NBPAIRS, outfp);

    fprintf(outfp, format_identifier, settype(S_H_DNA));
    fprintf(outfp, pairs_comment);
    for (c=1; c<NBPAIRS+1; c++)
      display_array(stackdH_D[c]+1,NBPAIRS,NBPAIRS, outfp);

    fprintf(outfp, format_identifier, settype(MMH_DNA));
    { int i,k;
    for (k=1; k<NBPAIRS+1; k++)
      for (i=0; i<5; i++)
        display_array(mismatchH37_D[k][i],5,5, outfp);
    }

    fprintf(outfp, format_identifier, settype(MMH_H_DNA));
    { int i,k;
    for (k=1; k<NBPAIRS+1; k++)
      for (i=0; i<5; i++)
        display_array(mismatchHdH_D[k][i],5,5, outfp);
    }

    fprintf(outfp, format_identifier, settype(MMI_DNA));
    { int i,k;
    for (k=1; k<NBPAIRS+1; k++)
      for (i=0; i<5; i++)
        display_array(mismatchI37_D[k][i],5,5, outfp);
    }

    fprintf(outfp, format_identifier, settype(MMI_H_DNA));
    { int i,k;
    for (k=1; k<NBPAIRS+1; k++)
      for (i=0; i<5; i++)
        display_array(mismatchIdH_D[k][i],5,5, outfp);
    }

    fprintf(outfp, format_identifier, settype(MMI1N_DNA));
    { int i,k;
    for (k=1; k<NBPAIRS+1; k++)
      for (i=0; i<5; i++)
        display_array(mismatch1nI37_D[k][i],5,5, outfp);
    }

    fprintf(outfp, format_identifier, settype(MMI1N_H_DNA));
    { int i,k;
    for (k=1; k<NBPAIRS+1; k++)
      for (i=0; i<5; i++)
        display_array(mismatch1nIdH_D[k][i],5,5, outfp);
    }

    fprintf(outfp, format_identifier, settype(MMI23_DNA));
    { int i,k;
    for (k=1; k<NBPAIRS+1; k++)
      for (i=0; i<5; i++)
        display_array(mismatch23I37_D[k][i],5,5, outfp);
    }

    fprintf(outfp, format_identifier, settype(MMI23_H_DNA));
    { int i,k;
    for (k=1; k<NBPAIRS+1; k++)
      for (i=0; i<5; i++)
        display_array(mismatch23IdH_D[k][i],5,5, outfp);
    }

    fprintf(outfp, format_identifier, settype(MMM_DNA));
    { int i,k;
    for (k=1; k<NBPAIRS+1; k++)
      for (i=0; i<5; i++)
        display_array(mismatchM37_D[k][i],5,5, outfp);
    }

    fprintf(outfp, format_identifier, settype(MMM_H_DNA));
    { int i,k;
    for (k=1; k<NBPAIRS+1; k++)
      for (i=0; i<5; i++)
        display_array(mismatchMdH_D[k][i],5,5, outfp);
    }

    fprintf(outfp, format_identifier, settype(MME_DNA));
    { int i,k;
    for (k=1; k<NBPAIRS+1; k++)
      for (i=0; i<5; i++)
        display_array(mismatchExt37_D[k][i],5,5, outfp);
    }

    fprintf(outfp, format_identifier, settype(MME_H_DNA));
    { int i,k;
    for (k=1; k<NBPAIRS+1; k++)
      for (i=0; i<5; i++)
        display_array(mismatchExtdH_D[k][i],5,5, outfp);
    }

    fprintf(outfp, format_identifier, settype(D5_DNA));
    fprintf(outfp, bases_comment);
    for (c=1; c<NBPAIRS+1; c++)
      display_array(dangle5_37_D[c], 5, 5, outfp);

    fprintf(outfp, format_identifier, settype(D5_H_DNA));
    fprintf(outfp, bases_comment);
    for (c=1; c<NBPAIRS+1; c++)
      display_array(dangle5_dH_D[c], 5, 5, outfp);

    fprintf(outfp, format_identifier, settype(D3_DNA));
    fprintf(outfp, bases_comment);
    for (c=1; c<NBPAIRS+1; c++)
      display_array(dangle3_37_D[c], 5, 5, outfp);

    fprintf(outfp, format_identifier, settype(D3_H_DNA));
    fprintf(outfp, bases_comment);
    for (c=1; c<NBPAIRS+1; c++)
      display_array(dangle3_dH_D[c], 5, 5, outfp);


    /* dont print "no pair" entries for interior loop arrays */
    fprintf(outfp, format_identifier, settype(INT11_DNA));
    { int i,k,l;
    for (k=1; k<NBPAIRS+1; k++)
      for (l=1; l<NBPAIRS+1; l++) {
        fprintf(outfp, format_pp, pnames[k], pnames[l]);
        for (i=0; i<5; i++)
          display_array(int11_37_D[k][l][i], 5, 5, outfp);
      }
    }

    fprintf(outfp, format_identifier, settype(INT11_H_DNA));
    { int i,k,l;
    for (k=1; k<NBPAIRS+1; k++)
      for (l=1; l<NBPAIRS+1; l++) {
        fprintf(outfp, format_pp, pnames[k], pnames[l]);
        for (i=0; i<5; i++)
          display_array(int11_dH_D[k][l][i],5,5, outfp);
      }
    }

    fprintf(outfp, format_identifier, settype(INT21_DNA));
    { int p1, p2, i, j;
    for (p1=1; p1<NBPAIRS+1; p1++)
      for (p2=1; p2<NBPAIRS+1; p2++)
        for (i=0; i<5; i++) {
          fprintf(outfp, format_pbp,
                  pnames[p1], bnames[i], pnames[p2]);
          for (j=0; j<5; j++)
            display_array(int21_37_D[p1][p2][i][j],5,5, outfp);
        }
    }

    fprintf(outfp, format_identifier, settype(INT21_H_DNA));
    { int p1, p2, i, j;
    for (p1=1; p1<NBPAIRS+1; p1++)
      for (p2=1; p2<NBPAIRS+1; p2++)
        for (i=0; i<5; i++) {
          fprintf(outfp, format_pbp,
                  pnames[p1], bnames[i], pnames[p2]);
          for (j=0; j<5; j++)
            display_array(int21_dH_D[p1][p2][i][j],5,5, outfp);
        }
    }

    fprintf(outfp, format_identifier, settype(INT22_DNA));
    { int p1, p2, i, j, k;
    for (p1=1; p1<NBPAIRS; p1++)
      for (p2=1; p2<NBPAIRS; p2++)
        for (i=1; i<5; i++)
          for (j=1; j<5; j++) {
            fprintf(outfp, format_pbbp,
                    pnames[p1], bnames[i], bnames[j], pnames[p2]);
            for (k=1; k<5; k++)
              display_array(int22_37_D[p1][p2][i][j][k]+1,4,5, outfp);
          }
    }

    fprintf(outfp, format_identifier, settype(INT22_H_DNA));
    { int p1, p2, i, j, k;
    for (p1=1; p1<NBPAIRS; p1++)
      for (p2=1; p2<NBPAIRS; p2++)
        for (i=1; i<5; i++)
          for (j=1; j<5; j++) {
            fprintf(outfp, format_pbbp,
                    pnames[p1], bnames[i], bnames[j], pnames[p2]);
            for (k=1; k<5; k++)
              display_array(int22_dH_D[p1][p2][i][j][k]+1,4,5, outfp);
          }
    }

    fprintf(outfp, format_identifier, settype(HP_DNA));
    display_array(hairpin37_D, 31, 10, outfp);

    fprintf(outfp, format_identifier, settype(HP_H_DNA));
    display_array(hairpindH_D, 31, 10, outfp);

    fprintf(outfp, format_identifier, settype(B_DNA));
    display_array(bulge37_D, 31, 10, outfp);

    fprintf(outfp, format_identifier, settype(B_H_DNA));
    display_array(bulgedH_D, 31, 10, outfp);

    fprintf(outfp, format_identifier, settype(IL_DNA));
    display_array(internal_loop37_D, 31, 10, outfp);

    fprintf(outfp, format_identifier, settype(IL_H_DNA));
    display_array(internal_loopdH_D, 31, 10, outfp);

    fprintf(outfp, format_identifier, settype(ML_DNA));
    fprintf(outfp,"/* F = cu*n_unpaired + cc + ci*loop_degree (+TermAU) */\n");
    fprintf(outfp,"/*\t    cu\t cu_dH\t    cc\t cc_dH\t    ci\t ci_dH  */\n");
    fprintf(outfp,"\t%6d\t%6d\t%6d\t%6d\t%6d\t%6d\n", ML_BASE37_D, ML_BASEdH_D, ML_closing37_D, ML_closingdH_D, ML_intern37_D, ML_interndH_D);

    fprintf(outfp, format_identifier, settype(NIN_DNA));
    fprintf(outfp,"/* Ninio = MIN(max, m*|n1-n2| */\n"
                "/*\t    m\t  m_dH     max  */\n"
                "\t%6d\t%6d\t%6d\n", ninio37_D, niniodH_D, MAX_NINIO_D);

    fprintf(outfp, format_identifier, settype(MISC_DNA));
    fprintf(outfp,"/* all parameters are pairs of 'energy enthalpy' */\n");
    fprintf(outfp,"/*    DuplexInit     TerminalAU      LXC */\n");
    fprintf(outfp,"   %6d %6d %6d  %6d %3.6f %6d\n", DuplexInit37_D, DuplexInitdH_D, TerminalAU37_D, TerminalAUdH_D, lxc37_D, 0);

    fprintf(outfp, format_identifier, settype(HEX_DNA));
    for (c=0; c< strlen(Hexaloops_D)/9; c++)
      fprintf(outfp,"\t%.8s %6d %6d\n", Hexaloops_D+c*9, Hexaloop37_D[c], HexaloopdH_D[c]);

    fprintf(outfp, format_identifier, settype(TL_DNA));
    for (c=0; c< strlen(Tetraloops_D)/7; c++)
      fprintf(outfp,"\t%.6s %6d %6d\n", Tetraloops_D+c*7, Tetraloop37_D[c], TetraloopdH_D[c]);

    fprintf(outfp, format_identifier, settype(TRI_DNA));
    for (c=0; c< strlen(Triloops_D)/6; c++)
      fprintf(outfp,"\t%.5s %6d %6d\n", Triloops_D+c*6, Triloop37_D[c], TriloopdH_D[c]);
  }

  if(options & VRNA_PARAMETER_FILE_HYBRID){
    fprintf(outfp, format_identifier, settype(S_HYBRID));
    fprintf(outfp, pairs_comment);
    for (c=1; c<NBPAIRS+1; c++)
      display_array(stack37_RD[c]+1,NBPAIRS,NBPAIRS, outfp);

    fprintf(outfp, format_identifier, settype(S_H_HYBRID));
    fprintf(outfp, pairs_comment);
    for (c=1; c<NBPAIRS+1; c++)
      display_array(stackdH_RD[c]+1,NBPAIRS,NBPAIRS, outfp);

    fprintf(outfp, format_identifier, settype(MMH_HYBRID));
    { int i,k;
    for (k=1; k<NBPAIRS+1; k++)
      for (i=0; i<5; i++)
        display_array(mismatchH37_RD[k][i],5,5, outfp);
    }

    fprintf(outfp, format_identifier, settype(MMH_H_HYBRID));
    { int i,k;
    for (k=1; k<NBPAIRS+1; k++)
      for (i=0; i<5; i++)
        display_array(mismatchHdH_RD[k][i],5,5, outfp);
    }

    fprintf(outfp, format_identifier, settype(MMI_HYBRID));
    { int i,k;
    for (k=1; k<NBPAIRS+1; k++)
      for (i=0; i<5; i++)
        display_array(mismatchI37_RD[k][i],5,5, outfp);
    }

    fprintf(outfp, format_identifier, settype(MMI_H_HYBRID));
    { int i,k;
    for (k=1; k<NBPAIRS+1; k++)
      for (i=0; i<5; i++)
        display_array(mismatchIdH_RD[k][i],5,5, outfp);
    }

    fprintf(outfp, format_identifier, settype(MMI1N_HYBRID));
    { int i,k;
    for (k=1; k<NBPAIRS+1; k++)
      for (i=0; i<5; i++)
        display_array(mismatch1nI37_RD[k][i],5,5, outfp);
    }

    fprintf(outfp, format_identifier, settype(MMI1N_H_HYBRID));
    { int i,k;
    for (k=1; k<NBPAIRS+1; k++)
      for (i=0; i<5; i++)
        display_array(mismatch1nIdH_RD[k][i],5,5, outfp);
    }

    fprintf(outfp, format_identifier, settype(MMI23_HYBRID));
    { int i,k;
    for (k=1; k<NBPAIRS+1; k++)
      for (i=0; i<5; i++)
        display_array(mismatch23I37_RD[k][i],5,5, outfp);
    }

    fprintf(outfp, format_identifier, settype(MMI23_H_HYBRID));
    { int i,k;
    for (k=1; k<NBPAIRS+1; k++)
      for (i=0; i<5; i++)
        display_array(mismatch23IdH_RD[k][i],5,5, outfp);
    }

    fprintf(outfp, format_identifier, settype(MMM_HYBRID));
    { int i,k;
    for (k=1; k<NBPAIRS+1; k++)
      for (i=0; i<5; i++)
        display_array(mismatchM37_RD[k][i],5,5, outfp);
    }

    fprintf(outfp, format_identifier, settype(MMM_H_HYBRID));
    { int i,k;
    for (k=1; k<NBPAIRS+1; k++)
      for (i=0; i<5; i++)
        display_array(mismatchMdH_RD[k][i],5,5, outfp);
    }

    fprintf(outfp, format_identifier, settype(MME_HYBRID));
    { int i,k;
    for (k=1; k<NBPAIRS+1; k++)
      for (i=0; i<5; i++)
        display_array(mismatchExt37_RD[k][i],5,5, outfp);
    }

    fprintf(outfp, format_identifier, settype(MME_H_HYBRID));
    { int i,k;
    for (k=1; k<NBPAIRS+1; k++)
      for (i=0; i<5; i++)
        display_array(mismatchExtdH_RD[k][i],5,5, outfp);
    }

    fprintf(outfp, format_identifier, settype(D5_HYBRID));
    fprintf(outfp, bases_comment);
    for (c=1; c<NBPAIRS+1; c++)
      display_array(dangle5_37_RD[c], 5, 5, outfp);

    fprintf(outfp, format_identifier, settype(D5_H_HYBRID));
    fprintf(outfp, bases_comment);
    for (c=1; c<NBPAIRS+1; c++)
      display_array(dangle5_dH_RD[c], 5, 5, outfp);

    fprintf(outfp, format_identifier, settype(D3_HYBRID));
    fprintf(outfp, bases_comment);
    for (c=1; c<NBPAIRS+1; c++)
      display_array(dangle3_37_RD[c], 5, 5, outfp);

    fprintf(outfp, format_identifier, settype(D3_H_HYBRID));
    fprintf(outfp, bases_comment);
    for (c=1; c<NBPAIRS+1; c++)
      display_array(dangle3_dH_RD[c], 5, 5, outfp);


    /* dont print "no pair" entries for interior loop arrays */
    fprintf(outfp, format_identifier, settype(INT11_HYBRID));
    { int i,k,l;
    for (k=1; k<NBPAIRS+1; k++)
      for (l=1; l<NBPAIRS+1; l++) {
        fprintf(outfp, format_pp, pnames[k], pnames[l]);
        for (i=0; i<5; i++)
          display_array(int11_37_RD[k][l][i], 5, 5, outfp);
      }
    }

    fprintf(outfp, format_identifier, settype(INT11_H_HYBRID));
    { int i,k,l;
    for (k=1; k<NBPAIRS+1; k++)
      for (l=1; l<NBPAIRS+1; l++) {
        fprintf(outfp, format_pp, pnames[k], pnames[l]);
        for (i=0; i<5; i++)
          display_array(int11_dH_RD[k][l][i],5,5, outfp);
      }
    }

    fprintf(outfp, format_identifier, settype(INT21_HYBRID));
    { int p1, p2, i, j;
    for (p1=1; p1<NBPAIRS+1; p1++)
      for (p2=1; p2<NBPAIRS+1; p2++)
        for (i=0; i<5; i++) {
          fprintf(outfp, format_pbp,
                  pnames[p1], bnames[i], pnames[p2]);
          for (j=0; j<5; j++)
            display_array(int21_37_RD[p1][p2][i][j],5,5, outfp);
        }
    }

    fprintf(outfp, format_identifier, settype(INT21_H_HYBRID));
    { int p1, p2, i, j;
    for (p1=1; p1<NBPAIRS+1; p1++)
      for (p2=1; p2<NBPAIRS+1; p2++)
        for (i=0; i<5; i++) {
          fprintf(outfp, format_pbp,
                  pnames[p1], bnames[i], pnames[p2]);
          for (j=0; j<5; j++)
            display_array(int21_dH_RD[p1][p2][i][j],5,5, outfp);
        }
    }

    fprintf(outfp, format_identifier, settype(INT22_HYBRID));
    { int p1, p2, i, j, k;
    for (p1=1; p1<NBPAIRS; p1++)
      for (p2=1; p2<NBPAIRS; p2++)
        for (i=1; i<5; i++)
          for (j=1; j<5; j++) {
            fprintf(outfp, format_pbbp,
                    pnames[p1], bnames[i], bnames[j], pnames[p2]);
            for (k=1; k<5; k++)
              display_array(int22_37_RD[p1][p2][i][j][k]+1,4,5, outfp);
          }
    }

    fprintf(outfp, format_identifier, settype(INT22_H_HYBRID));
    { int p1, p2, i, j, k;
    for (p1=1; p1<NBPAIRS; p1++)
      for (p2=1; p2<NBPAIRS; p2++)
        for (i=1; i<5; i++)
          for (j=1; j<5; j++) {
            fprintf(outfp, format_pbbp,
                    pnames[p1], bnames[i], bnames[j], pnames[p2]);
            for (k=1; k<5; k++)
              display_array(int22_dH_RD[p1][p2][i][j][k]+1,4,5, outfp);
          }
    }

    fprintf(outfp, format_identifier, settype(HP_HYBRID));
    display_array(hairpin37_RD, 31, 10, outfp);

    fprintf(outfp, format_identifier, settype(HP_H_HYBRID));
    display_array(hairpindH_RD, 31, 10, outfp);

    fprintf(outfp, format_identifier, settype(B_HYBRID));
    display_array(bulge37_RD, 31, 10, outfp);

    fprintf(outfp, format_identifier, settype(B_H_HYBRID));
    display_array(bulgedH_RD, 31, 10, outfp);

    fprintf(outfp, format_identifier, settype(IL_HYBRID));
    display_array(internal_loop37_RD, 31, 10, outfp);

    fprintf(outfp, format_identifier, settype(IL_H_HYBRID));
    display_array(internal_loopdH_RD, 31, 10, outfp);

    fprintf(outfp, format_identifier, settype(ML_HYBRID));
    fprintf(outfp,"/* F = cu*n_unpaired + cc + ci*loop_degree (+TermAU) */\n");
    fprintf(outfp,"/*\t    cu\t cu_dH\t    cc\t cc_dH\t    ci\t ci_dH  */\n");
    fprintf(outfp,"\t%6d\t%6d\t%6d\t%6d\t%6d\t%6d\n", ML_BASE37_RD, ML_BASEdH_RD, ML_closing37_RD, ML_closingdH_RD, ML_intern37_RD, ML_interndH_RD);

    fprintf(outfp, format_identifier, settype(NIN_HYBRID));
    fprintf(outfp,"/* Ninio = MIN(max, m*|n1-n2| */\n"
                "/*\t    m\t  m_dH     max  */\n"
                "\t%6d\t%6d\t%6d\n", ninio37_RD, niniodH_RD, MAX_NINIO_RD);

    fprintf(outfp, format_identifier, settype(MISC_HYBRID));
    fprintf(outfp,"/* all parameters are pairs of 'energy enthalpy' */\n");
    fprintf(outfp,"/*    DuplexInit     TerminalAU      LXC */\n");
    fprintf(outfp,"   %6d %6d %6d  %6d %3.6f %6d\n", DuplexInit37_RD, DuplexInitdH_RD, TerminalAU37_RD, TerminalAUdH_RD, lxc37_RD, 0);

    fprintf(outfp, format_identifier, settype(HEX_HYBRID));
    for (c=0; c< strlen(Hexaloops_RD)/9; c++)
      fprintf(outfp,"\t%.8s %6d %6d\n", Hexaloops_RD+c*9, Hexaloop37_RD[c], HexaloopdH_RD[c]);

    fprintf(outfp, format_identifier, settype(TL_HYBRID));
    for (c=0; c< strlen(Tetraloops_RD)/7; c++)
      fprintf(outfp,"\t%.6s %6d %6d\n", Tetraloops_RD+c*7, Tetraloop37_RD[c], TetraloopdH_RD[c]);

    fprintf(outfp, format_identifier, settype(TRI_HYBRID));
    for (c=0; c< strlen(Triloops_RD)/6; c++)
      fprintf(outfp,"\t%.5s %6d %6d\n", Triloops_RD+c*6, Triloop37_RD[c], TriloopdH_RD[c]);
  }

  fprintf(outfp, format_identifier, settype(QUIT));
  fclose(outfp);
}

PRIVATE void check_symmetry(void) {
  int i,j,k,l;

  for (i=0; i<=NBPAIRS; i++)
    for (j=0; j<=NBPAIRS; j++)
      if (stack37[i][j] != stack37[j][i])
        fprintf(stderr, "WARNING: stacking energies not symmetric\n");

  for (i=0; i<=NBPAIRS; i++)
    for (j=0; j<=NBPAIRS; j++)
      if (stackdH[i][j] != stackdH[j][i])
        fprintf(stderr, "WARNING: stacking enthalpies not symmetric\n");


  /* interior 1x1 loops */
  for (i=0; i<=NBPAIRS; i++)
    for (j=0; j<=NBPAIRS; j++)
      for (k=0; k<5; k++)
        for (l=0; l<5; l++)
          if (int11_37[i][j][k][l] != int11_37[j][i][l][k])
            fprintf(stderr, "WARNING: int11 energies not symmetric (%d,%d,%d,%d) (%d vs. %d)\n", i, j, k, l, int11_37[i][j][k][l], int11_37[j][i][l][k]);

  for (i=0; i<=NBPAIRS; i++)
    for (j=0; j<=NBPAIRS; j++)
      for (k=0; k<5; k++)
        for (l=0; l<5; l++)
          if (int11_dH[i][j][k][l] != int11_dH[j][i][l][k])
            fprintf(stderr, "WARNING: int11 enthalpies not symmetric\n");

  /* interior 2x2 loops */
  for (i=0; i<=NBPAIRS; i++)
    for (j=0; j<=NBPAIRS; j++)
      for (k=0; k<5; k++)
        for (l=0; l<5; l++) {
          int m,n;
          for (m=0; m<5; m++)
            for (n=0; n<5; n++)
              if (int22_37[i][j][k][l][m][n] != int22_37[j][i][m][n][k][l])
                fprintf(stderr, "WARNING: int22 energies not symmetric\n");
        }

  for (i=0; i<=NBPAIRS; i++)
    for (j=0; j<=NBPAIRS; j++)
      for (k=0; k<5; k++)
        for (l=0; l<5; l++) {
          int m,n;
          for (m=0; m<5; m++)
            for (n=0; n<5; n++)
              if (int22_dH[i][j][k][l][m][n] != int22_dH[j][i][m][n][k][l])
                fprintf(stderr, "WARNING: int22 enthalpies not symmetric: %d %d %d %d %d %d\n", i,j,k,l,m,n);
        }
}

/* update nonstandard nucleotide/basepair involved contributions for int22 */
PRIVATE void update_nst(int array[NBPAIRS+1][NBPAIRS+1][5][5][5][5]){
  int    i, j, k, l, m, n;
  int max, max2, max3, max4, max5, max6;

  /* get maxima for one nonstandard nucleotide */
  for (i=1; i<NBPAIRS; i++){
    for (j=1; j<NBPAIRS; j++){
      for (k=1; k<5; k++){
        for (l=1; l<5; l++){
          for (m=1; m<5; m++){
            max = max2 = max3 = max4 = -INF; /* max of {CGAU} */
            for(n=1;n<5;n++){
              max   = MAX2(max,   array[i][j][k][l][m][n]);
              max2  = MAX2(max2,  array[i][j][k][l][n][m]);
              max3  = MAX2(max3,  array[i][j][k][n][l][m]);
              max4  = MAX2(max4,  array[i][j][n][k][l][m]);
            }
            array[i][j][k][l][m][0] = max;
            array[i][j][k][l][0][m] = max2;
            array[i][j][k][0][l][m] = max3;
            array[i][j][0][k][l][m] = max4;
          }
        }
      }
    }
  }
  /* get maxima for two nonstandard nucleotides */
  for (i=1; i<NBPAIRS; i++){
    for (j=1; j<NBPAIRS; j++){
      for (k=1; k<5; k++){
        for (l=1; l<5; l++){
          max = max2 = max3 = max4 = max5 = max6 = -INF; /* max of {CGAU} */
          for (m=1; m<5; m++){
            max   = MAX2(max,   array[i][j][k][l][m][0]);
            max2  = MAX2(max2,  array[i][j][k][m][0][l]);
            max3  = MAX2(max3,  array[i][j][m][0][k][l]);
            max4  = MAX2(max4,  array[i][j][0][k][l][m]);
            max5  = MAX2(max5,  array[i][j][0][k][m][l]);
            max6  = MAX2(max6,  array[i][j][k][0][l][m]);
          }
          array[i][j][k][l][0][0] = max;
          array[i][j][k][0][0][l] = max2;
          array[i][j][0][0][k][l] = max3;
          array[i][j][k][0][l][0] = max6;
          array[i][j][0][k][0][l] = max5;
          array[i][j][0][k][l][0] = max4;
        }
      }
    }
  }
  /* get maxima for three nonstandard nucleotides */
  for (i=1; i<NBPAIRS; i++){
    for (j=1; j<NBPAIRS; j++){
      for (k=1; k<5; k++){
        max = max2 = max3 = max4 = -INF; /* max of {CGAU} */
        for (l=1; l<5; l++){
          /* should be arbitrary where index l resides in last 3 possible locations */
          max   = MAX2(max,   array[i][j][k][l][0][0]);
          max2  = MAX2(max2,  array[i][j][0][k][l][0]);
          max3  = MAX2(max3,  array[i][j][0][0][k][l]);
          max4  = MAX2(max4,  array[i][j][0][0][l][k]);
        }
        array[i][j][k][0][0][0] = max;
        array[i][j][0][k][0][0] = max2;
        array[i][j][0][0][k][0] = max3;
        array[i][j][0][0][0][k] = max4;
      }
    }
  }
  /* get maxima for 4 nonstandard nucleotides */
  for (i=1; i<NBPAIRS; i++){
    for (j=1; j<NBPAIRS; j++){
      max = -INF; /* max of {CGAU} */
      for (k=1; k<5; k++){
        max   = MAX2(max,   array[i][j][k][0][0][0]);
      }
      array[i][j][0][0][0][0] = max;
    }
  }

  /* now compute contributions for nonstandard base pairs ... */
  /* first, 1 nonstandard bp */
  for (i=1; i<NBPAIRS; i++){
    for (k=0; k<5; k++){
      for (l=0; l<5; l++){
        for (m=0; m<5; m++){
          for(n=0;n<5;n++){
            max = max2 = -INF;
            for(j=1;j<NBPAIRS;j++){
              max   = MAX2(max, array[i][j][k][l][m][n]);
              max2  = MAX2(max2, array[j][i][k][l][m][n]);
            }
            array[i][NBPAIRS][k][l][m][n] = max;
            array[NBPAIRS][i][k][l][m][n] = max2;
          }
        }
      }
    }
  }

  /* now 2 nst base pairs */
  for (k=0; k<5; k++){
    for (l=0; l<5; l++){
      for (m=0; m<5; m++){
        for(n=0;n<5;n++){
          max = -INF;
          for(j=1;j<NBPAIRS;j++){
            max   = MAX2(max, array[NBPAIRS][j][k][l][m][n]);
          }
          array[NBPAIRS][NBPAIRS][k][l][m][n] = max;
        }
      }
    }
  }

}
